package dev.fitko.fitconnect.core.validation;

import com.fasterxml.jackson.core.JacksonException;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.networknt.schema.JsonSchema;
import com.networknt.schema.JsonSchemaFactory;
import com.networknt.schema.SpecVersion;
import com.networknt.schema.ValidationMessage;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.util.Base64;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.DataIntegrityException;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.util.Strings;
import dev.fitko.fitconnect.jwkvalidator.JWKValidator;
import dev.fitko.fitconnect.jwkvalidator.JWKValidatorBuilder;
import dev.fitko.fitconnect.jwkvalidator.exceptions.JWKValidationException;
import dev.fitko.fitconnect.jwkvalidator.exceptions.LogLevel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.nio.charset.StandardCharsets;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.jwkvalidator.JWKValidator.withX5CValidation;
import static dev.fitko.fitconnect.jwkvalidator.JWKValidator.withoutX5CValidation;

public class DefaultValidationService implements ValidationService {

    private static final Logger LOGGER = LoggerFactory.getLogger(DefaultValidationService.class);

    private static final ObjectMapper MAPPER = new ObjectMapper();

    private final MessageDigestService messageDigestService;
    private final ApplicationConfig config;

    public DefaultValidationService(final MessageDigestService messageDigestService, final ApplicationConfig config) {
        this.messageDigestService = messageDigestService;
        this.config = config;
    }

    @Override
    public ValidationResult validatePublicKey(final RSAKey publicKey) {
        try {
            validateKey(publicKey);
        } catch (final JWKValidationException | CertificateEncodingException e) {
            return ValidationResult.error(e);
        }
        return ValidationResult.ok();
    }

    @Override
    public ValidationResult validateMetadataSchema(final String jsonSchema, final Metadata metadata) {
        try {
            final String metadataJson = MAPPER.writeValueAsString(metadata);
            final JsonNode schemaNode = MAPPER.readTree(jsonSchema);
            final JsonNode inputNode = MAPPER.readTree(metadataJson);
            final JsonSchemaFactory factory = getJsonSchemaFactory();
            final JsonSchema schema = factory.getSchema(schemaNode);
            return returnValidationResult(schema.validate(inputNode));
        } catch (final JsonProcessingException e) {
            return ValidationResult.error(e);
        }
    }

    @Override
    public ValidationResult validateIntegrity(final String originalHexHash, final byte[] data) {
        try {
            final byte[] originalHash = this.messageDigestService.fromHexString(originalHexHash);
            final boolean hashesAreNotEqual = !this.messageDigestService.verify(originalHash, data);
            if (hashesAreNotEqual) {
                return ValidationResult.error(new DataIntegrityException("Hash sum of transmitted data does not equal the hash of the sender."));
            }
            return ValidationResult.ok();
        } catch (final IllegalArgumentException | NullPointerException e) {
            return ValidationResult.error(e);
        }
    }

    @Override
    public ValidationResult validateJsonFormat(final String json) {
        try {
            MAPPER.readTree(json);
        } catch (final JacksonException e) {
            return ValidationResult.error(e);
        }
        return ValidationResult.ok();
    }

    @Override
    public ValidationResult validateXmlFormat(final String xml) {
        try {
            final XMLReader xmlReader = getXmlReader();
            xmlReader.parse(new InputSource(new ByteArrayInputStream(xml.getBytes(StandardCharsets.UTF_8))));
        } catch (final ParserConfigurationException | IOException | SAXException e) {
            return ValidationResult.error(e);
        }
        return ValidationResult.ok();
    }

       private ValidationResult returnValidationResult(final Set<ValidationMessage> errors) {
        if (errors.isEmpty()) {
            return ValidationResult.ok();
        }
        return ValidationResult.error(new ValidationException(errorsToSingleString(errors)));
    }

    private void validateKey(final RSAKey publicKey) throws JWKValidationException, CertificateEncodingException {
        if (this.config.getCurrentEnvironment().isAllowInsecurePublicKey()) {
            validateWithoutCertChain(publicKey);
        } else {
            validateCertChain(publicKey);
        }
    }

    private void validateWithX509AndProxy(final List<String> pemCerts, final RSAKey publicKey) throws JWKValidationException {
        final var proxy = new Proxy(Proxy.Type.HTTP, new InetSocketAddress(this.config.getHttpProxyHost(), this.config.getHttpProxyPort()));
        LOGGER.info("Using proxy {} for key validation", proxy);
        addLogLevel(withX5CValidation()
                .withProxy(proxy)
                .withRootCertificatesAsPEM(pemCerts))
                .validate(publicKey);
    }


    private void validateWithX509AndWithoutProxy(final List<String> pemCerts, final RSAKey publicKey) throws JWKValidationException {
        LOGGER.info("No proxy set for validation");
        addLogLevel(withX5CValidation()
                .withoutProxy()
                .withRootCertificatesAsPEM(pemCerts))
                .validate(publicKey);
    }

    private JWKValidator addLogLevel(final JWKValidatorBuilder.JWKValidatorX5CErrorHandling validator) {
        if (validateSilent()) {
            return validator.withoutThrowingExceptions().withErrorLogLevel(LogLevel.WARN).build();
        } else {
            return validator.withThrowingExceptions().withErrorLogLevel(LogLevel.ERROR).build();
        }
    }

    private void validateWithoutCertChain(final RSAKey publicKey) throws JWKValidationException {
        LOGGER.info("Validating public key without XC5 certificate chain");
        withoutX5CValidation().withErrorLogLevel(validateSilent() ? LogLevel.WARN : LogLevel.ERROR).build().validate(publicKey);
    }

    private void validateCertChain(final RSAKey publicKey) throws JWKValidationException, CertificateEncodingException {
        final List<String> pemCerts = getPemCerts(publicKey.getParsedX509CertChain());
        LOGGER.info("Validation public key with XC5 certificate chain checks");
        if (isProxySet()) {
            validateWithX509AndProxy(pemCerts, publicKey);
        } else {
            validateWithX509AndWithoutProxy(pemCerts, publicKey);
        }
    }

    private List<String> getPemCerts(final List<X509Certificate> certChain) throws CertificateEncodingException {
        final List<String> pemCerts = new ArrayList<>();
        for (final X509Certificate cert : certChain) {
            pemCerts.add(getEncodedString(cert));
        }
        return pemCerts;
    }

    private String getEncodedString(final X509Certificate x509Certificate) throws CertificateEncodingException {
        return Base64.encode(x509Certificate.getEncoded()).toString();
    }

    private boolean isProxySet() {
        return this.config.getHttpProxyPort() != null && Strings.isNotNullOrEmpty(this.config.getHttpProxyHost());
    }
    
    private boolean validateSilent() {
        return this.config.getCurrentEnvironment().isAllowInsecurePublicKey();
    }

    private String errorsToSingleString(final Set<ValidationMessage> errors) {
        return errors.stream()
                .map(ValidationMessage::getMessage)
                .collect(Collectors.joining("\n"));
    }

    private static JsonSchemaFactory getJsonSchemaFactory() {
        return JsonSchemaFactory.getInstance(SpecVersion.VersionFlag.V202012);
    }

    private static XMLReader getXmlReader() throws ParserConfigurationException, SAXException {
        final SAXParserFactory saxParserFactory = SAXParserFactory.newInstance();
        saxParserFactory.setFeature("http://apache.org/xml/features/disallow-doctype-decl", true);
        return saxParserFactory.newSAXParser().getXMLReader();
    }

}
